*
*  X3GENPK.PRG
*  Return a unique PK character string from a GENPK table record
*  IF GENPK fails, return value is .NULL.
*
*  Copyright (c) 1996  (Public Domain) MaxTech, Inc. 
*                      MaxTech, Inc.
*                      8721C Plantation Lane, Suite 302
*                      Manassas, VA  20110
*                      703-369-1759
*  Author:  Drew Speedie
*
*
*  NOTE:  Requires a contained "generated PK" table 
*         with table name GENERATEPK and the
*         following fields:
*     GPK_PK            C(40)  TableName for the generated PK 
*     GPK_CurrentNumber I(4) 
*     GPK_CurrentAlpha  C(5) 
*
*
*  NOTE:  Works with contained tables only --
*         the passed tcGPK_PK is checked against
*         the passed tcGenPKDatabaseName to see if it is a valid
*         TableName (if it is an open alias(), 
*         the alias is checked via CURSORGETPROP()
*         to make sure we're using the TableName,
*         not some current alias)
*
*
*  NOTE:  If the GENERATEPK table is already open, 
*         optimistic buffering is expected and 
*         will be set/reset. 
*         If the GENERATEPK table is not already open,
*         it will be opened, optimistic table
*         buffering set, and left open
*
*  NOTE:  If you're using character PKs, when
*         you reach "zzzzz", the limit (somewhere between 5-6 gig) 
*         on the the incrementing routine, the RETURN value 
*         is always .NULL.
*
*  NOTE:  If you're using numeric (Integer) PKs, when
*         you reach the 2147483647 (2 gig) limit on the
*         integer field capacity, the RETURN value 
*         is always .NULL.
*
*  NOTE:  If you're using character PKs, the following ASCII
*         characters never appear in the generated PK:
*            33 ! (exclamation point)
*            34 " (double-quotes character delimiter)
*            40 ' (single-quote character delimiter)
*            91 [ (left bracket)
*            93 ] (right bracket)
*           122 | (pipe character)
*           123 { (left curly brace)
*           124 } (right curly brace)
*
*
*  USAGE: 
*  lcID = X3GENPK("CUSTOMERS","VM") 
*  IF isnull(lcID)
*    * take appropriate action on failure
*   ELSE
*    lcID = padl(lcID,"0",lenc(CUSTOMERS.Cus_PK))  
*    replace <PKField> with lcID
*  ENDIF
*
*
*  lParameters 
*          tcGPK_PK (R) Type of PK to be generated 
*                       Corresponds to GENERATEPK.GPK_PK
*   tcGENPKDatabase (R) Database containing GENERATEPK
*        tcGENPKTag (O) GENPK index tag for this operation
*                       Defaults to "GPK_PK"
*        tnAttempts (O) Number of attempts to try
*                       Defaults to 1000
*                       If <=0, try forever
*         tlNumeric (O) Indicates whether the RETURN value
*                         is to be used as a numeric/integer,
*                         in which case you don't get the
*                         rollover behavior from numerics to
*                         alphas at 99999.
*                       NOTE:  if the Primary() key tag
*                         expression is a Numeric or Integer value,
*                         you get tlNumeric=.t. behavior WITHOUT
*                         having to explicitly pass it as such.
*                         You only have to pass tlNumeric=.t. 
*                         when, for some reason, you are using
*                         Character PKs and you want to handle the
*                         data conversion yourself or do not 
*                         want the rollover behavior on reaching
*                         99999.
*                         So the default/not passed behavior:
*                         1- if the PK of tcGPK_PK is an expression
*                            of data type Character, the RETURN 
*                            value is a character string, either
*                            the alltrim(str()) of a numeric or,
*                            on reaching "99999", a unique 5-character
*                            string
*                         2- if the PK of tcGPK_PK is an expression
*                            of data type Integer/Numeric, the RETURN
*                            value is a number
*                       
*                      
lparameters tcGPK_PK, ;
            tcGENPKDatabaseName, ;
            tcGENPKTag, ;
            tnAttempts, ;
            tlNumeric
local luRetVal, ;
      lnAttemptsSoFar, ;
      lcAlias, ;
      lcAliasDatabase, ;
      lcSetDatabase, ;
      lnBufferMode, ;
      llClosePKTable, ;
      llBogus

luRetVal = space(0)
luRetVal = .NULL.

IF type("tcGPK_PK")#"C" OR empty(tcGPK_PK)
  * not passed/passed improperly:
  return luRetVal
 ELSE
  tcGPK_PK = upper(alltrim(tcGPK_PK)) 
ENDIF
IF type("tcGENPKDatabaseName")#"C" ;
     OR empty(tcGENPKDatabaseName)
  * not passed/passed improperly:
  return luRetVal
 ELSE
  tcGENPKDatabaseName = upper(alltrim(tcGENPKDatabaseName))
  IF occurs("\",tcGENPKDatabasename)>0
    tcGENPKDatabaseName = substrc(tcGENPKDatabaseName, ;
                                 at_c("\",tcGENPKDatabaseName,occurs("\",tcGENPKDatabaseName))+1)
    tcGENPKDatabasename = leftc(tcGENPKDatabaseName,at_c(".",tcGENPKDatabaseName)-1)
  ENDIF
ENDIF

* save the current alias() information:
lcAlias         = alias()
lcAliasDatabase = space(0)
IF !empty(lcAlias)
  lcAliasDatabase = cursorgetprop("DATABASE")   &&& empty() if a free table
ENDIF
lcSetDatabase = set("DATABASE")

IF dbused(tcGENPKDatabaseName)
 ELSE
  open data (tcGENPKDatabaseName)
ENDIF  
set Database to (tcGENPKDatabaseName)

local llGoodTableName
IF type("tcGENPKTag") # "C" ;
     OR empty(tcGENPKTag)
  * not passed/passed improperly, assume tag: 
  tcGENPKTag = "GPK_PK" 
ENDIF

IF !used("GENERATEPK")
  use GENERATEPK in 0
  lnBufferMode = 1
  =cursorsetprop("BUFFERING",5,"GENERATEPK")
 ELSE
  select GENERATEPK
  lnBufferMode = cursorgetprop("BUFFERING","GENERATEPK")
  IF lnBufferMode # 5
    =cursorsetprop("BUFFERING",5,"GENERATEPK")
  ENDIF
ENDIF

IF used(tcGPK_PK)
  * 
  * make sure we know the TableName for the passed
  * tcGPK_PK (it could have been passed as an alias)
  *
  tcGPK_PK = cursorgetprop("SourceName",tcGPK_PK)
  llGoodTableName = .t.
  llClosePKTable  = .f.
 ELSE
  * 
  * make sure the passed tcGPK_PK is
  * a TableName for a table that exists
  * in the current database (tcGenPKDatabaseName)
  *
  llGoodTableName = indbc(tcGPK_PK,"TABLE")
  llClosePKTable  = .t.
  IF llGoodTableName
    use (tcGPK_PK) in 0 
  ENDIF
ENDIF

IF !llGoodTableName
  return .NULL.
ENDIF

=seek(tcGPK_PK, ;
      "GENERATEPK", ;
      tcGENPKTag)

* if there's not one on file, add it now
IF !found("GENERATEPK")
  insert into GENERATEPK (GPK_PK, ;
                          GPK_CurrentNumber, ;
                          GPK_CurrentAlpha) ;
                          values ;
                          (tcGPK_PK, ;
                           0, ;
                           space(0)) 
  DO WHILE .t.
    *
    *  we've set optimistic table buffering
    *
    IF tableupdate(.t.,.f.,"GENERATEPK")
      exit
    ENDIF
  ENDDO
ENDIF

*
*  determine how long to try 
*
IF type("tnAttempts")="N"
  IF tnAttempts <= 0
    tnAttempts = 0
  ENDIF 
 ELSE
  tnAttempts = 1000
ENDIF

*
* are we looking for an alpha
* or numeric return value?
*
local lcReturnValueDataType

IF type("tlNumeric") = "L" AND tlNumeric
  *
  *  numeric has been explicitly requested
  *
  lcReturnValueDataType = "N"
 ELSE
  *
  *  determine the data type of the PK for
  *  the tcGPK_PK table ("N" or "C")
  *
  select (tcGPK_PK)
  lcReturnValueDataType = dbgetprop(tcGPK_PK,"TABLE","PRIMARYKEY")
  lcReturnValueDataType = sys(14,tagno(lcReturnValueDataType))
  lcReturnValueDataType = type("eval(lcReturnValueDataType)")
  select GENERATEPK
ENDIF

lnAttemptsSoFar = 0
llBogus = .f.
DO WHILE .t.
  local llNumber, llAlpha
  IF lcReturnValueDataType = "N" ;
       OR (lcReturnValueDataType = "C" AND GPK_CurrentNumber < 99999)
    *
    *  explicitly requested Numeric/Integer
    *  or 
    *  Character, GPK_CurrentNumber < 99999
    *
    IF GPK_CurrentNumber = 2147483647    &&& Integer field capacity!!
      llBogus = .t.
     ELSE
      replace GPK_CurrentNumber with ; 
           GPK_CurrentNumber + 1 ;
           IN GENERATEPK
      llNumber = .t.
    ENDIF
   ELSE
    *
    *  Character, GPK_CurrentNumber = 99999
    *
    IF GPK_CurrentAlpha = replicate(chr(122),lenc(GPK_CurrentAlpha))
      * Alpha field limit
      llBogus = .t.
     ELSE
      replace GPK_CurrentAlpha with ;
           IncrementAlphaNumeric(GPK_CurrentAlpha) ;
           IN GENERATEPK
      llAlpha = .t.
    ENDIF
  ENDIF
  IF llBogus
    exit
  ENDIF
  IF tableupdate(.t.,.f.,"GENERATEPK")
    DO CASE
      CASE lcReturnValueDataType = "C" AND llNumber 
        luRetVal = alltrim(str(GPK_CurrentNumber))
      CASE lcReturnValueDataType = "C" AND llAlpha
        luRetVal = alltrim(GPK_CurrentAlpha)
      OTHERWISE numeric
        luRetVal = GPK_CurrentNumber
    ENDCASE    
    exit
   ELSE
    *
    *  dump the buffer and start over
    *
    =tablerevert(.t.,"GENERATEPK")
  ENDIF  
  lnAttemptsSoFar = lnAttemptsSoFar + 1
  IF lnAttemptsSoFar = tnAttempts 
    * give up:
    exit
  ENDIF
ENDDO

IF "2" $ getfldstate(-1,"GENERATEPK")
  *
  *  we had to bail out of the above loop
  *  without a successful TABLEUPDATE()
  *
  =tablerevert(.t.,"GENERATEPK")
ENDIF

IF lnBufferMode # cursorgetprop("BUFFERING","GENERATEPK")
  *
  *  we made sure here that we're using
  *  optimistic table buffering.  If
  *  GENERATEPK was already open in some
  *  other buffer mode, set it back
  *
  =cursorsetprop("BUFFERING",lnBufferMode,"GENERATEPK")
ENDIF

if .f.
IF empty(lcAliasDatabase)
  * lcAlias is a free table
 ELSE
  set database to (lcAliasDatabase)
ENDIF
endif

IF llClosePKTable
  use in (tcGPK_PK)
ENDIF

IF empty(lcAlias)
  select 0
 ELSE
  select (lcAlias)
ENDIF  

IF empty(lcSetDatabase)
  set database to
 ELSE 
  set database to (lcSetDatabase)
ENDIF  

return luRetVal



PROCEDURE IncrementAlphaNumeric
*
*  On reaching a PK of "99999", PK generation
*  switches over to an arbitrary string of
*  ASCII characters from 35 to 122, skipping
*  numbers and character delimiters: '[]
* 
*  Which means the following characters do not
*  appear in generated PKs:
*     33 !
*     34 "
*     40 '
*     91 [
*     93 ]
*    122 |
*    123 {
*    124 }
*   
lparameters tcPKString
local lcString
IF empty(tcPKString)
  *
  *  first value past "99999"
  *
  lcString = replicate(chr(35),5) 
 ELSE
  * increment the alphanumeric
  *  ASCII values from 35 - 122
  *
  local lcOne, lcTwo, lcThree, lcFour, lcFive
  lcOne   = leftc(tcPKString,1)
  lcTwo   = substrc(tcPKString,2,1)
  lcThree = substrc(tcPKString,3,1)
  lcFour  = substrc(tcPKString,4,1)
  lcFive  = rightc(tcPKString,1)
  IF asc(lcFive) = 122
    lcFive = chr(35)
    IF asc(lcFour) = 122
      lcFour = chr(35)
      IF asc(lcThree) = 122
        lcThree = chr(35)
        IF asc(lcTwo) = 122
          lcTwo = chr(35)
          lcOne = chr(asc(lcOne)+1)
          * skip numbers and character delimiters:
          lcOne = SkipChars(lcOne)
         ELSE
          lcTwo = chr(asc(lcTwo)+1)
          * skip numbers and character delimiters:
          lcTwo = SkipChars(lcTwo) 
        ENDIF asc(lcTwo) = 122
       ELSE
        lcThree = chr(asc(lcThree)+1)
        * skip numbers and character delimiters:
        lcThree = SkipChars(lcThree)
      ENDIF asc(lcThree) = 122
     ELSE
      lcFour = chr(asc(lcFour)+1)
      * skip numbers and character delimiters:
      lcFour = SkipChars(lcFour)
    ENDIF asc(lcFour) = 122
   ELSE
    lcFive = chr(asc(lcFive)+1)   
    * skip numbers and character delimiters:
    lcFive = SkipChars(lcFive) 
  ENDIF asc(lcFive) = 122 
  lcString = lcOne + lcTwo + lcThree + lcFour + lcFive
ENDIF
return lcString



PROCEDURE SkipChars
lparameters lcASCII
local lcReturnValue
lcReturnValue = lcASCII
DO CASE
  CASE asc(lcASCII) = 48  
    * skip numbers
    lcReturnValue = chr(58)
  CASE asc(lcASCII) = 39
    * skip the ' character delimiter
    lcReturnValue = chr(40)
  CASE asc(lcASCII) = 91
    * skip the [ character delimiter
    lcReturnValue = chr(92)
  CASE asc(lcASCII) = 93
    * skip the ] character delimiter
    lcReturnValue = chr(94)
ENDCASE
return lcReturnValue



PROCEDURE keywords
* Primary Keys 
return